      subroutine bicgstab(ia,n,ngx,ngy,nxpml,nypml,xl,yl, 
     &                    a,b,x,ap,inump,itdata,
     &                    imgrid,rmgrid,mmax,ngall0,ngall1,nipre,nrpre,
     &                    ipm,ipre,rpre,wpre,w,work)
c     
      implicit none 
c 
      integer     ia(2),ip(5),inump(9),ipm(4),igrid  
      integer     j,n,ngx,ngy,iw,jw,nxpml,nypml   
      integer     iter,maxiter,iprec,ipsol 
      integer     nmatvec,itdata(5)    
c
      complex*16  a(*),ap(*),b(n),x(n)  
      complex*16  w(n,7) 
      complex*16  rho0,rho1,alpha,beta,omega,r0v,ts,tt 
      complex*16  herpr 
c      
      real*8      xl,yl 
      real*8      bb,rr,rr0,rel,tol,resdata(2)    
      real*8      t1,t2,t3,t4,tinit,tend,time(5)  
      real*8      dnorm2  
c      
      integer     imgrid(13),mmax,ngall0,ngall1,nipre,nrpre 
      real*8      rmgrid(3)        
      integer     ipre(nipre) 
      complex*16  rpre(nrpre),wpre(nrpre) 
      complex*16  wrelax 
c
      complex*16  work(n+2,3) 
c      
c     ******************************************************************
c
      do j = 1,5
         time(j) = 0.d0
      enddo
      
      call cputime(tinit)
      iter    = 0
      nmatvec = 0
      ngx     = ia(1)
      ngy     = ia(2)
      maxiter = inump(2)
      tol     = 10.**(inump(3))
      iprec   = inump(4)
      ipsol   = inump(5) 
      wrelax  = cmplx(rmgrid(2),rmgrid(3)) 
c
      open(20,file='residual.out')
c
c     Calculate BB=(B,B). If BB = 0, then singular problem.
c      
      bb = dnorm2(b,n)
      if (bb .lt. 1.d-32) then
          write(*,*) 'Singular problem. Program is terminated.'
          return
      endif      
c
c     ******************************************************************    
c     (2) Initialization
c
c     Compute W5 = A*X
c
      call cputime(t1)
      call mat2d9p(ngx,ngy,x,w(1,5),a,work(1,1))
      call cputime(t2)
      time(2) = time(2) + (t2-t1)
      nmatvec = nmatvec + 1

C     Compute the initial residual R0 = W5 = B-A*X = B-W5 
C     Set R = R0 = W1, and W2 = W3 = 0
C
      do j=1,n
         w(j,5) = b(j) - w(j,5)
         w(j,1) = w(j,5)
         w(j,2) = (0.d0,0.d0)
         w(j,3) = (0.d0,0.d0)
      enddo

C     Check the accuracy of the initial solution
      rr  = dnorm2(w(1,1),n)
c      rs  = dnorm2(s,n)
      if (rr .le. 1.d-32) then
         write(*,*) 'Solution is already obtained without iteration.'
         write(*,*) 'Program is terminated'
         return
      else
         rr0 = rr
         rel = rr/rr0
         write(*,100)  iter,nmatvec,rr,rel
         write(20,100) iter,nmatvec,rr,rel
      endif     
C     
      rho0  = (1.d0,0.d0)
      alpha = (1.d0,0.d0)
      omega = (1.d0,0.d0)
      
c     ******************************************************************
c     Prepare the preconditioner
      call pmgrid(n,ngx,ngy,ap,ia,xl,yl,
     &            imgrid,rmgrid,mmax,ngall0,ngall1,nipre,nrpre,    
     &            wpre,rpre,ipre,ip,work)   
C
C     ******************************************************************    
C     (3) BiCGStab Iteration
C
      call cputime(tend)
      time(1) = time(1) + (tend-tinit)
      time(5) = time(5) + (tend-tinit)
      
      call cputime(tinit)
      do while (iter .lt. maxiter .and. rel .gt. tol)
         
         iter = iter + 1
C
C        Compute RHO1 = (R0,R) = (W5,W1)
C        Compute BETA = (RHO1/RHO0)/(ALPHA/OMEGA)
C        Update  RHO0 = RHO1
C
         rho1 = herpr(w(1,5),w(1,1),n)
         beta = (rho1/rho0)*(alpha/omega)
         rho0 = rho1
C
C        Compute W3=W1+BETA*(W3-OMEGA*W2)
C
         do j = 1,n
            w(j,3) = w(j,1) + beta*(w(j,3) - omega*w(j,2))
         enddo
C     
C        Preconditioner solve: M*W4 = W3
C
         call cputime(t3)
         if (iprec .eq. 0) then
            do j = 1,n
               w(j,4) = w(j,3)
            enddo
         else
            call pmg2d(n,xl,yl,w(1,3),w(1,4),ap,wrelax,ia,wpre,rpre,
     &                 ipre,work) 
         endif
         call cputime(t4)
         time(3) = time(3) + (t4-t3)
c     
c        Compute W2 = A*W4
c
      	 call cputime(t1)
         call mat2d9p(ngx,ngy,w(1,4),w(1,2),a,work(1,1))
         call cputime(t2)
         time(2) = time(2) + (t2-t1)
         nmatvec = nmatvec + 1
c
c        Compute ALPHA = RHO0/(W5,W2)
c
         r0v = herpr(w(1,5),w(1,2),n)
         alpha = rho0/r0v
c     
c        Compute W6 = W1 - ALPHA*W2
c         
         do j = 1,n
            w(j,6) = w(j,1)-alpha*w(j,2)
         enddo
c
c        Preconditioner solve: M*W1 = W6
c
         call cputime(t3)
         if (iprec .eq. 0) then
            do j = 1,n
              w(j,1) = w(j,6)
            enddo
         else
            call pmg2d(n,xl,yl,w(1,6),w(1,1),ap,wrelax,ia,wpre,rpre,
     &                 ipre,work) 
         endif
         call cputime(t4)
         time(3) = time(3) + (t4-t3)
c 
c        Compute W7 = A*W1
c
         call cputime(t1)
         call mat2d9p(ngx,ngy,w(1,1),w(1,7),a,work(1,1))
         call cputime(t2)
         time(2) = time(2) + (t2-t1)
         nmatvec = nmatvec + 1
c
c        Compute OMEGA = (W7,W6)/(W7,W7)
c
         ts = herpr(w(1,7),w(1,6),n)
         tt = herpr(w(1,7),w(1,7),n)
         omega = ts/tt
c     
c        Update the residual X  = X + ALPHA*W4 + OMEGA*W1
c        Update the residual W1 = W6 - OMEGA*W7
c
         do j=1,n
            x(j)   = x(j) + alpha*w(j,4) + omega*w(j,1)
            w(j,1) = w(j,6) - omega*w(j,7)
         enddo
c
c        Convergence criterion
c
         rr  = dnorm2(w(1,1),n)
         rel = rr/rr0
c
         write(*,100)  iter,nmatvec,rr,rel
         write(20,100) iter,nmatvec,rr,rel
c
      enddo
c
      call cputime(tend)
      time(4) = time(4) + (tend-tinit)
      time(5) = time(5) + (tend-tinit)
c
      itdata(1) = iter
      itdata(2) = nmatvec
      resdata(1) = rr
      resdata(2) = rel
c
      call prntdata(ia,n,nxpml,nypml,xl,yl,inump,itdata,time,x,resdata)
c       
  100 format(' ',i7,' ',i7,'   ',d14.8,'   ',d14.8)
c      
      return
      end
