      SUBROUTINE GRMG2D(NGX,NGY,M,ICOMM, 
     &                  NX,NY, 
     &                  INDX,
     &                  NGALL0, 
     &                  IERR, 
     &                  NPX,NPY,IPX,IPY, 
     &                  IX0,IY0,
     &                  NEIGX,NEIGY, 
     &                  NEIGXC,NEIGYC, 
     &                  IXG,IYG, 
     &                  IPPX,IPPY, 
     &                  NPPX,NPPY)

C     ******************************************************************
C
C     The subroutine GRMG2D constructs the level of multigrid, based  on
C     data from the finest grid level. The data required are
C
C        NGX   : # of gridpoints in x-direction at the finest level
C        NGY   : # of gridpoints in y-direction at the finest level
C        M     : # of MG level
C
C     These are user-defineable.
C
C     The level is numbered in such a way that the finer grid is identi-
C     fied by larger level number. For example, M is the designated num-
C     ber for the finest level, while 1 is for the coarsest one.
C
C     The subroutine then returns 
C
C     Variables definition:
C
C        NX(M) : # of gridpoints in x-direction at every level. NX is  a
C                one-dimension array.
C        NY(M) : # of gridpoints in x-direction at every level. NY is  a  
C                one-dimension array.
C
C     ******************************************************************

      IMPLICIT   REAL*8 (A-H,O-Z)
      INTEGER    NX(M),NY(M),INDX(M),IX0(M),IY0(M) 
      INTEGER    IXG(M),IYG(M),
     &           IPPX(M),IPPY(M),NPPX(M),NPPY(M),
     &           NEIGX(2,M),NEIGY(2,M),NEIGXC(2,M),NEIGYC(2,M)
      INTEGER    NXA(0:255),NYA(0:255) 

      ICOMM = 0
C
      IPE = IPX + IPY*NPX
C
      
      DO K = M,1,-1
      
C     Compute # of gridpoints at every level. For level < M, the # of
C     grid points is determined by doubling the gridsize.

         IF (K .EQ. M) THEN
            NX(K) = NGX
            NY(K) = NGY
         ELSE
            NX(K) = (NX(K+1) - IX0(K+1) + 2)/2
            NY(K) = (NY(K+1) - IY0(K+1) + 2)/2
         ENDIF
C         
         NX0 = NX(K)
         NY0 = NY(K)
         CALL COMMGB2D(NPX,NPY,IPX,IPY,NX0,NY0,
     &                 IXG0,IYG0,ICOMM)
                  
         IXG(K) = IXG0
         IYG(K) = IYG0
         IX0(K) = MOD(IXG0+1,2) + 1
         IY0(K) = MOD(IYG0+1,2) + 1
         
         CALL COMMAS2D(NPX,NPY,IPX,IPY,
     &                 NX0,NY0,NXA(0),NYA(0),ICOMM)
     
         NXX     = 0
         NPPX(K) = 0
         IPPX(K) = 0
         DO IX = 0,NPX-1
            NXX = NXX + NXA(IX)
            IF (NXA(IX) .NE. 0) NPPX(K) = NPPX(K) + 1
         ENDDO
         
         NYY     = 0
         NPPY(k) = 0
         DO IY = 0,NPY-1
            NYY = NYY + NYA(IY)
            IF (NYA(IY) .NE. 0) NPPY(K) = NPPY(K) + 1
         ENDDO
         
         IF (MIN(NXX,NYY) .LE. 0) THEN
            IERR = 21
            RETURN
         ENDIF
C
         IF ((NX0*NY0) .EQ. 0) THEN
            IPPX(K) = -1
            IPPY(K) = -1
         ELSE
            IPPX(k) = 0
            DO IX = 0,IPX-1
               IF (NXA(IX) .NE. 0) IPPX(K) = IPPX(K) + 1
            ENDDO
            IPPY(K) = 0
            DO IY = 0,IPY-1
               IF (NYA(IY) .NE. 0) IPPY(K) = IPPY(K) + 1
            ENDDO
         ENDIF
C
         NEIGX(1,K)  = -1
         NEIGXC(1,K) = -1
         NEIGX(2,K)  = -1
         NEIGXC(2,K) = -1
         NEIGY(1,K)  = -1
         NEIGYC(1,K) = -1
         NEIGY(2,K)  = -1
         NEIGYC(2,K) = -1
C
         IF (NX(K) .GT. 0 .AND. NY(K) .GT. 0) THEN
            DO IPX0 = IPX-1,0,-1
               IF (NXA(IPX0) .GT. 0) THEN
                  NEIGX(1,K) = IPX0 + IPY*NPX
                  GOTO 100
               ENDIF
            ENDDO
 100        CONTINUE
            DO IPX0 = IPX+1,NPX-1
               IF (NXA(IPX0) .GT. 0) THEN
                  NEIGX(2,K) = IPX0 + IPY*NPX
                  GOTO 110
               ENDIF
            ENDDO
 110        CONTINUE
            DO IPY0 = IPY-1,0,-1
               IF (NYA(IPY0) .GT. 0) THEN
                  NEIGY(1,K) = IPX + IPY0*NPX
                  GOTO 120
               ENDIF
            ENDDO
 120        CONTINUE
            DO IPY0 = IPY+1,NPY-1
               IF (NYA(IPY0) .GT. 0) THEN
                  NEIGY(2,K) = IPX + IPY0*NPX
                  GOTO 130
               ENDIF
            ENDDO
 130        CONTINUE
            DO IXOF = 1,NPX
               IX = MOD(IPX-IXOF+NPX,NPX)
               IF (NXA(IX) .GT. 0) THEN
                  NEIGXC(1,K) = IX + IPY*NPX
                  GOTO 140
               ENDIF
            ENDDO
 140        CONTINUE
            DO IXOF = 1,NPX
               IX = MOD(IPX+IXOF,NPX)
               IF (NXA(IX) .GT. 0) THEN
                  NEIGXC(2,K) = IX + IPY*NPX
                  GOTO 150
               ENDIF
            ENDDO
 150        CONTINUE
            DO IYOF = 1,NPY
               IY = MOD(IPY-IYOF+NPY,NPY)
               IF (NYA(IY) .GT. 0) THEN
                  NEIGYC(1,K) = IPX + IY*NPX
                  GOTO 160
               ENDIF
            ENDDO
 160        CONTINUE
            DO IYOF = 1,NPY
               IY = MOD(IPY+IYOF,NPY)
               IF (NYA(IY) .GT. 0) THEN
                  NEIGYC(2,K) = IPX + IY*NPX
                  GOTO 170
               ENDIF
            ENDDO
 170        CONTINUE
         ENDIF
C    
      ENDDO 
C
C     Make INDX
C
      INDX(1) = 1
c      WRITE(*,*) nx(1),ny(1),indx(1)
      DO K = 2,M
         INDX(K) = INDX(K-1) + NX(K-1)*NY(K-1)
c         WRITE(*,*) K, NX(K), NY(K),INDX(K)
      ENDDO
      NGALL0 = INDX(M)-1
c      WRITE(*,*) NGALL0
c      
      RETURN
      END
