C     General
C     =======
C     The HELM3D solves the Helmholtz equation numerically. The core  of
C     the solver is iterative solver based on Bi-CGStab. Some precondit-
C     ioning techniques are incorporated and is user-defined. By precon-
C     ditioning, the performance of Bi-CGStab can be enhanced.
C
C     HELM3D solves the Helmholtz equation in a square or box computati-
C     onal domain with equidistant grids. The  equation  is  discretized
C     using Finite Difference Methods. Discretizations currently availa-
C     ble are:
C
C        -  2nd order FDM
C        -  4th order FDM
C
C     Boundary conditions can be chosen for either Dirichlet, Neumann,or
C     Sommerfeld type.
C
C     For the iterative methods, BiCGStab is mainly used. The  algorithm
C     follows immediately from H.A.van der Vorst (SIAM J. Sci. Comput.).
C     For efficiency, only right preconditioning is implemented.   Using
C     only right preconditioning removes extra two preconditioner solves
C     necessary for left preconditioning. 
C
C     Preconditioners currently available are:
C
C        -  ILU based on A (the original matrix)
C        -  Shifted Laplace operator

      program helm2d
c
      implicit none
c      
c     To initialize, we need to set a vector length, which be used
c     to define the size of matrices and vectors involved in this code.
c     The vector length is called nmax and is chosen such that
c
c           nmax > ngx*ngy
c
c     if ngx, ngy the number of gridpoints in x and y direction, resp.
c
      integer     nmax 
      parameter   (nmax = 300000)
c
      integer     irhs,ibca(2,2),ibcp(2,2),icase 
      integer     itsol,iprec,ipsol,itmax,itol,insol,inprec 
      integer     itstart,itrunc,iprtsol,iprtres,idisc,igrid  
      integer     ixy,ix,iy,ngx,ngy,ngxp,ngyp,ngxd,ngyd 
      real*8      xl,yl,xs,ys,wnum,fac,wpml,cdl,coeffp(2)   
      complex*16  a(9*nmax),ap(9*nmax),b(nmax),x(nmax) 
c  
c     multigrid parameter
c
      integer     npx,npy,m,nit,nr1,nr2 
      integer     icyc,md,imin,iac,mr,mj 
      integer     ids,iswp 
      real*8      epsac,wr(2) 
c      
      integer     mmax,nipremax,nrpremax,nwpremax,nall0,nall1 
      parameter   (mmax = 20,
     &             nall0 = 0.5*nmax,
     &             nall1 = nmax + nall0,
     &             nipremax = 19*mmax + 10,
     &             nrpremax = nall0*9 + nall0, !nall0*22 + 4*nall1,
     &             nwpremax = nall0*9 + 4*nall1) !nall0*22 + 4*nall1)
c      
      integer     ipre(nipremax) 
      complex*16  rpre(nrpremax),wpre(nwpremax) 
c
      complex*16  w(7*nmax),work(nmax*3),s(nmax,3),v(nmax,1) 
c     
c     OPENING MESSAGE
      call pgopen     
c     
      open(1,file='input2d.dat')
      read(1,*)
      read(1,*)
      read(1,*) xl,yl,wnum,fac 
      read(1,*)
      read(1,*)
      read(1,*) xs,ys
      read(1,*)
      read(1,*)
      read(1,*) ngxp,ngyp,wpml 
      read(1,*)
      read(1,*)
      read(1,*) idisc,icase,cdl
      read(1,*)
      read(1,*)
      read(1,*) ibca(1,1),ibca(1,2),ibca(2,1),ibca(2,2)
      read(1,*)
      read(1,*) 
      read(1,*) itsol,iprec,ipsol,insol,inprec 
      read(1,*) 
      read(1,*)
      read(1,*) coeffp(1),coeffp(2)
      read(1,*)
      read(1,*)
      read(1,*) ibcp(1,1),ibcp(1,2),ibcp(2,1),ibcp(2,2)
      read(1,*)
      read(1,*)
      read(1,*) itmax,itol,itstart,itrunc 
      read(1,*)
      read(1,*)
      read(1,*) npx,npy,m
      read(1,*)
      read(1,*)
      read(1,*) nit,nr1,nr2
      read(1,*)
      read(1,*)
      read(1,*) icyc,md,mj
      read(1,*)
      read(1,*)
      read(1,*) imin,iac,mr,epsac
      read(1,*)
      read(1,*)
      read(1,*) ids,iswp,wr(1),wr(2)
      read(1,*)
      read(1,*)
      read(1,*) iprtsol,iprtres
c
      if (ngxp*ngyp .gt. nmax) then
         write(*,*) 'Grid size is too large.'
         stop
      endif
c
c     Prepare the matrices and vector
c
      if (idisc .eq. 1) then        
         call mka5pb1(xl,yl,ngxp,ngyp,wnum,fac,ibca,a,icase)
         call makeblo1(ngxp,ngyp,xl,yl,xs,ys,b)
         call mkp5pb1(xl,yl,ngxp,ngyp,wnum,fac,coeffp,ibcp,ap,icase)
         ngx = ngxp
         ngy = ngyp
         igrid = 0
      elseif (idisc .eq. 2) then        
         call mka5pb2(xl,yl,ngxp,ngyp,wnum,fac,ibca,a,icase)
         call makeblo2(ngxp,ngyp,xl,yl,xs,ys,b)
         call mkp5pb2(xl,yl,ngxp,ngyp,wnum,fac,coeffp,ibcp,ap,icase)
         ngx = ngxp+2
         ngy = ngyp+2
         igrid = 1
      elseif (idisc .eq. 3) then
         ngxd = wpml*ngxp
	 ngyd = wpml*ngyp
	 call mka9jabl(xl,yl,ngxp,ngyp,ngxd,ngyd,wnum,fac,ibca,cdl,
     &                 a,icase)
         call makebhodl(ngxp,ngyp,ngxd,ngyd,xl,yl,xs,ys,b)
	 call mkp5pabl(xl,yl,ngxp,ngyp,ngxd,ngyd,wnum,fac,coeffp,ibcp,
     &	               cdl,ap,icase)
         ngx = ngxp + 2*ngxd
	 ngy = ngyp + 2*ngyd
	 igrid = 2
      else
         write(*,*) 'No matrices are built. Program is terminated.'
	 stop 
      endif
c
      ixy = 1
      do iy = 1,ngy
         do ix = 1,ngx
            x(ixy) = (0.d0,0.d0)
            ixy = ixy + 1
         enddo
      enddo
c
c     Start the method
c
      call driver(xl,yl,ngx,ngy,ngxd,ngyd,a,b,x,ap,
     &            itsol,iprec,ipsol,insol,inprec,
     &            itmax,itol,itstart,itrunc,igrid,iprtsol,iprtres,
     &            npx,npy,m,nit,nr1,nr2,icyc,md,imin,iac,mr,epsac,
     &            mj,ids,iswp,wr,
     &            ipre,rpre,wpre,w,work,s,v)
c      
      end
