c     This subroutine constructs and returns the matrix A obtained  from
c     2-nd order acccurate FD of the Helmholtz equation. The boundary is
c     of first order Sommerfeld type discretize using the 1-st order 
c     one-sided FD.
c 
      subroutine  mka5pb1(xl,yl,ngx,ngy,wreff,fac,ibc,a,icase)   
c  
      implicit     none
      integer      ix,iy,ngx,ngy,ibc(2,2),icase   
      complex*16   a(ngx,ngy,9) 
      complex*16   czero,cone,cidd 
      complex*16   omg 
      real*8       fac,omg_re,omg_im,wvnum 
      real*8       alpha,hx,hy,r1,r2,r3,r4,wreff,wdata 
      real*8       c1,c2,c3,c5,cho,gamm 
      real*8       xl,yl 
      real*8       dxw,nxw,sxw,dxe,nxe,sxe,dyn,nyn,syn,dys,nys,sys 
c     
      parameter      (czero = (0.d0,0.d0),
     &                cone  = (1.d0,0.d0),
     &                cidd  = (0.d0,1.d0)) 
c
c     make the matrix elements for 2-D grids
c      
      hx= xl/(dble(float(ngx+1)))
      hy= yl/(dble(float(ngy+1)))
c
      r1 =  1.d0/hy**2
      r2 =  1.d0/hx**2
      r3 =  2.d0/hy**2
      r4 =  2.d0/hx**2
c
      do iy = 1,ngy
         do ix = 1,ngx
            wvnum = wdata(ix,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
            omg = cmplx(1.d0,-fac)*wvnum**2
            a(ix,iy,1) =  czero
            a(ix,iy,2) = -r1*cone
            a(ix,iy,3) =  czero
            a(ix,iy,4) = -r2*cone
            a(ix,iy,5) =  (r3 + r4) - omg
            a(ix,iy,6) = -r2*cone
            a(ix,iy,7) =  czero
            a(ix,iy,8) = -r1*cone
            a(ix,iy,9) =  czero
        enddo
      enddo
c
c     Resolve the boundary conditions
c
c     West
      if (ibc(1,1) .eq. 0) then
         dxw = 0.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 1) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 2) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 1.d0
      endif
c     East
      if (ibc(1,2) .eq. 0) then
         dxe = 0.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 1) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 2) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 1.d0
      endif
c     South
      if (ibc(2,1) .eq. 0) then
         dys = 0.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 1) then
         dys = 1.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 2) then
         dys = 1.d0
         nys = 1.d0
         sys = 1.d0
      endif
c     North
      if (ibc(2,2) .eq. 0) then
         dyn = 0.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 1) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 2) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 1.d0
      endif    
c
c     Boundary condition 
c     N : North
      do ix = 1, ngx
         wvnum = wdata(ix,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase) 
         omg = cmplx(1.d0,-fac)*wvnum
         a(ix,ngy,5) = a(ix,ngy,5) - dyn/(nyn 
     &               + syn*cidd*omg*hy)/hy**2.d0
         a(ix,ngy,7) = 0.d0
         a(ix,ngy,8) = 0.d0
         a(ix,ngy,9) = 0.d0
      enddo
c     S : South
      do ix = 1, ngx
         wvnum = wdata(ix,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
         omg = cmplx(1.d0,-fac)*wvnum
         a(ix,1,5) = a(ix,1,5) - dys/(nys 
     &             + sys*cidd*omg*hy)/hy**2.d0
         a(ix,1,1) = 0.d0
         a(ix,1,2) = 0.d0
         a(ix,1,3) = 0.d0
      enddo
c     E : East
      do iy = 1, ngy
         wvnum = wdata(ngx+1,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
         omg = cmplx(1.d0,-fac)*wvnum
         a(ngx,iy,5) = a(ngx,iy,5) - dxe/(nxe 
     &               + sxe*cidd*omg*hx)/hx**2.d0
         a(ngx,iy,3) = 0.d0
         a(ngx,iy,6) = 0.d0
         a(ngx,iy,9) = 0.d0
      enddo
c     W : West
      do iy = 1, ngy
         wvnum = wdata(0,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
         omg = cmplx(1.d0,-fac)*wvnum
         a(1,iy,5) = a(1,iy,5) - dxw/(nxw 
     &             + sxw*cidd*omg*hx)/hx**2.d0
         a(1,iy,1) = 0.d0
         a(1,iy,4) = 0.d0
         a(1,iy,7) = 0.d0
      enddo
c      
      return
      end
