      subroutine mkp5pb2(xl,yl,ngx,ngy,wreff,fac,coeffp,ibc,a,icase)
      
c     This subroutine returns the preconditioner matrix A with the second
c     order boundary condition, discretized using central FD scheme.

      implicit     none
      integer      ix,iy,ngx,ngy,ibc(2,2),icase   
      complex*16   a(0:ngx+1,0:ngy+1,9) 
      complex*16   czero,cone,cidd 
      complex*16   omg,omg1,omg2  
      real*8       hx,hy,r1,r2,r3,r4 
      real*8       coeffp(2),fac,wreff,wdata,wvnum  
      real*8       xl,yl  
      real*8       dxw,nxw,sxw,dxe,nxe,sxe,dyn,nyn,syn,dys,nys,sys 
c      real*8       vdata(0:ngx+1,0:ngy+1) 
      real*8       pi 
     
      parameter      (czero = (0.d0,0.d0),
     &                cone  = (1.d0,0.d0),
     &                cidd  = (0.d0,1.d0)) 
c
      pi = 4.d0*atan(1.d0)
c     
c     make the matrix elements for 2-D grids
c
      hx= xl/(dble(float(ngx+1)))
      hy= yl/(dble(float(ngy+1)))
c
      r1 =  1.d0/hy**2
      r2 =  1.d0/hx**2
      r3 =  2.d0/hy**2
      r4 =  2.d0/hx**2
c
      do iy = 0,ngy+1
         do ix = 0,ngx+1
            wvnum = wdata(ix,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
            omg = cmplx(coeffp(1),-coeffp(2))*cmplx(1.d0,-fac)*wvnum**2
            a(ix,iy,1) =  czero
            a(ix,iy,2) = -r1*cone
            a(ix,iy,3) =  czero
            a(ix,iy,4) = -r2*cone
            a(ix,iy,5) =  (r3 + r4)*cone - omg
            a(ix,iy,6) = -r2*cone
            a(ix,iy,7) =  czero
            a(ix,iy,8) = -r1*cone 
            a(ix,iy,9) =  czero
         enddo
      enddo
c
c     Resolve the boundary conditions
c
c     West
      if (ibc(1,1) .eq. 0) then
         dxw = 0.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 1) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 2) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 1.d0
      endif
c     East
      if (ibc(1,2) .eq. 0) then
         dxe = 0.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 1) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 2) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 1.d0
      endif
c     South
      if (ibc(2,1) .eq. 0) then
         dys = 0.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 1) then
         dys = 1.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 2) then
         dys = 1.d0
         nys = 1.d0
         sys = 1.d0
      endif
c     North
      if (ibc(2,2) .eq. 0) then
         dyn = 0.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 1) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 2) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 1.d0
      endif
c
c     Boundary condition
c
c     North
      do ix = 1,ngx
         wvnum = wdata(ix,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c         wvnum = 2.d0*pi*wreff/vdata(ix,ngy+1)
         omg   = cmplx(1.d0,-fac)*wvnum
c         write(*,*) omg
         a(ix,ngy+1,1) =  czero
         a(ix,ngy+1,2) = -2.d0/hy**2
         a(ix,ngy+1,3) =  czero
         a(ix,ngy+1,4) =  a(ix,ngy+1,4) + cidd/(omg*hy)/hx**2
         a(ix,ngy+1,5) =  a(ix,ngy+1,5) 
     &                 +  (2.d0*cidd*(omg*hy - 1.d0/(omg*hx)))/hy**2
         a(ix,ngy+1,6) =  a(ix,ngy+1,6) + cidd/(omg*hy)/hx**2
         a(ix,ngy+1,7) =  czero
         a(ix,ngy+1,8) =  czero
         a(ix,ngy+1,9) =  czero
      enddo
c     South
      do ix = 1,ngx
         wvnum = wdata(ix,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c         wvnum = 2.d0*pi*wreff/vdata(ix,0)
         omg   = cmplx(1.d0,-fac)*wvnum
c         write(*,*) omg
         a(ix,0,1) =  czero
         a(ix,0,2) =  czero
         a(ix,0,3) =  czero
         a(ix,0,4) =  a(ix,0,4) + cidd/(omg*hy)/hx**2
         a(ix,0,5) =  a(ix,0,5) 
     &             + (2.d0*cidd*(omg*hy - 1.d0/(omg*hx)))/hy**2
         a(ix,0,6) =  a(ix,0,6) + cidd/(omg*hy)/hx**2
         a(ix,0,7) =  czero
         a(ix,0,8) = -2.d0/hy**2
         a(ix,0,9) =  czero
      enddo
c     East 
      do iy = 1,ngy
         wvnum = wdata(ngx+1,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c         wvnum = 2.d0*pi*wreff/vdata(ngx+1,iy)
         omg   = cmplx(1.d0,-fac)*wvnum
c         write(*,*) omg
         a(ngx+1,iy,1) =  czero
         a(ngx+1,iy,2) =  a(ngx+1,iy,2) + cidd/(omg*hx)/hy**2
         a(ngx+1,iy,3) =  czero
         a(ngx+1,iy,4) = -2.d0/hx**2
         a(ngx+1,iy,5) =  a(ngx+1,iy,5) 
     &                 + (2.d0*cidd*(omg*hx - 1.d0/(omg*hy)))/hx**2      
         a(ngx+1,iy,6) =  czero
         a(ngx+1,iy,7) =  czero
         a(ngx+1,iy,8) =  a(ngx+1,iy,8) + cidd/(omg*hx)/hy**2
         a(ngx+1,iy,9) =  czero
      enddo
c     West 
      do iy = 1,ngy
         wvnum = wdata(0,iy,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c         wvnum = 2.d0*pi*wreff/vdata(0,iy)
         omg   = cmplx(1.d0,-fac)*wvnum
c         write(*,*) omg
         a(0,iy,1) =  czero
         a(0,iy,2) =  a(0,iy,2) + cidd/(omg*hx)/hy**2
         a(0,iy,3) =  czero
         a(0,iy,4) =  czero
         a(0,iy,5) =  a(0,iy,5) 
     &             + (2.d0*cidd*(omg*hx - 1.d0/(omg*hy)))/hx**2
         a(0,iy,6) = -2.d0/hx**2
         a(0,iy,7) =  czero
         a(0,iy,8) =  a(0,iy,8) + cidd/(omg*hx)/hy**2
         a(0,iy,9) =  czero
      enddo
c      
c     Corner point
c      
c     South-West
c
      wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c      wvnum = 2.d0*pi*wreff/vdata(0,0)
      omg   = cmplx(1.d0,-fac)*wvnum
c      write(*,*) omg
      a(0,0,1) =  czero
      a(0,0,2) =  czero
      a(0,0,3) =  czero
      a(0,0,4) =  czero
      a(0,0,5) =  a(0,0,5) + 4.d0*cidd*omg*hx/hy**2
      a(0,0,6) = -2.d0/hx**2
      a(0,0,7) =  czero
      a(0,0,8) = -2.d0/hy**2
      a(0,0,9) =  czero
c
c     South-East
c     
      wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c      wvnum = 2.d0*pi*wreff/vdata(ngx+1,0)
      omg   = cmplx(1.d0,-fac)*wvnum
c      write(*,*) omg
      a(ngx+1,0,1) =  czero
      a(ngx+1,0,2) =  czero
      a(ngx+1,0,3) =  czero
      a(ngx+1,0,4) = -2.d0/hx**2
      a(ngx+1,0,5) =  a(ngx+1,0,5) + 4.d0*cidd*omg*hx/hy**2
      a(ngx+1,0,6) =  czero
      a(ngx+1,0,7) =  czero
      a(ngx+1,0,8) = -2.d0/hy**2
      a(ngx+1,0,9) =  czero
c
c     North-West
c
      wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c      wvnum = 2.d0*pi*wreff/vdata(0,ngy+1)
      omg   = cmplx(1.d0,-fac)*wvnum
c      write(*,*) omg
      a(0,ngy+1,1) =  czero
      a(0,ngy+1,2) = -2.d0/hy**2
      a(0,ngy+1,3) =  czero
      a(0,ngy+1,4) =  czero
      a(0,ngy+1,5) =  a(0,ngy+1,5) + 4.d0*cidd*omg*hx/hy**2
      a(0,ngy+1,6) = -2.d0/hx**2
      a(0,ngy+1,7) =  czero
      a(0,ngy+1,8) =  czero
      a(0,ngy+1,9) =  czero
c
c     North-East
c
      wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
c      wvnum = 2.d0*pi*wreff/vdata(ngx+1,ngy+1)
      omg   = cmplx(1.d0,-fac)*wvnum
c      write(*,*) omg
      a(ngx+1,ngy+1,1) =  czero
      a(ngx+1,ngy+1,2) = -2.d0/hy**2
      a(ngx+1,ngy+1,3) =  czero
      a(ngx+1,ngy+1,4) = -2.d0/hx**2
      a(ngx+1,ngy+1,5) =  a(ngx+1,ngy+1,5) + 4.d0*cidd*omg*hx/hy**2
      a(ngx+1,ngy+1,6) =  czero
      a(ngx+1,ngy+1,7) =  czero
      a(ngx+1,ngy+1,8) =  czero
      a(ngx+1,ngy+1,9) =  czero
c
      return
      end
