c     This subroutine returns the preconditioner M of the linear system
c
c              AM^{-1} z = b,  z = M x.
c
c     This matrix M is required for convergence acceleration of Krylov
c     subspace methods, which in this case Bi-CGSTAB. The preconditioner
c     has the following features:
c
c         (i) it is generated from the 5-point stencil FDM applied to the
c             shifted Laplace operator. For preconditioner, it is not
c             necessary that the same stencil as for the Helmholtz equation
c             is used. In fact, this is one of the advantage of shifted
c             Laplace preconditioner!
c
c        (ii) Since the absorption layer is used in the Helmholtz equation
c             (i.e. in A), the absorption layer is also used in M. The same
c             treatment as in A is also used in M.
c
c
      subroutine mkp5pabl(xl,yl,ngx,ngy,ngxd,ngyd,wreff,fac,coeff,ibc,
     &                    const,a,icase)
c
      integer    ix,iy,ngx,ngy,ngxd,ngyd,ngxt,ngyt,icase,ibc(2,2)   
      integer    dxw,nxw,sxw,dxe,nxe,sxe,dyn,nyn,syn,dys,nys,sys 
      real*8     hx,hy,hx2,hy2,xl,yl,xd,yd,coeff(2),c1,c2    
      real*8     ac,am,as,alf,const,fac,gamm,wreff,wvnum,wdata   
      complex*16 a(ngx+2*ngxd,ngy+2*ngyd,9),axx,ayy,acx,acy,omg,cidd 
      parameter  (cidd = (0.d0,1.d0))
c      
      c1 = coeff(1)
      c2 = coeff(2)
c
      hx  = xl/dble(float(ngx+1))
      hy  = yl/dble(float(ngy+1))
      hx2 = hx**2
      hy2 = hy**2
      xd  = hx*dble(float(ngxd))
      yd  = hy*dble(float(ngyd))
c      
      ngxt = ngx + 2*ngxd
      ngyt = ngy + 2*ngyd
c
      do iy = 1,ngyt
        if (iy .ge. 1 .and. iy .le. ngyd-1) then
          do ix = 1,ngxt
            if (ix .ge. 1 .and. ix .le. ngxd-1) then
c
              wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngxd-ix)*hx/xd)**2 
     &                    + const*((ngyd-iy)*hy/yd)**2
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0     
c
            elseif (ix .eq. ngxd) then
c
              wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)       
              alf   = fac + const*((ngyd-iy)*hy/yd)**2  
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0          
c         
            elseif (ix .ge. ngxd+1 .and. ix .le. ngxd+ngx) then
c
              wvnum = wdata(ix-ngxd,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngyd-iy)*hy/yd)**2
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .eq. ngxd+ngx+1) then
c
              wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)    
              alf   = fac + const*((ngyd-iy)*hy/yd)**2
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c                              
            elseif (ix .ge. ngxd+ngx+2 .and. ix .le. ngxt) then
c
              wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ix-(ngxd+ngx+1))*hx/xd)**2
     &                    + const*((ngyd-iy)*hy/yd)**2  
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            endif
          enddo
        elseif (iy .eq. ngyd) then
          do ix = 1,ngxt
            if (ix .ge. 1 .and. ix .le. ngxd-1) then
              wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngxd-ix)*hx/xd)**2     
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .eq. ngxd) then
c
              wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .ge. ngxd+1 .and. ix .le. ngxd+ngx) then
              wvnum = wdata(ix-ngxd,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac  
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .eq. ngxd+ngx+1) then
c
              wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac    
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .ge. ngxd+ngx+2 .and. ix .le. ngxt) then  
c
              wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ix-(ngxd+ngx+1))*hx/xd)**2     
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            endif
          enddo
        elseif (iy .ge. ngyd+1 .and. iy .le. ngyd+ngy) then
          do ix = 1,ngxt
            if (ix .ge. 1 .and. ix .le. ngxd-1) then
c            
              wvnum = wdata(0,iy-ngyd,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngxd-ix)*hx/xd)**2  
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .eq. ngxd) then
c
              wvnum = wdata(0,iy-ngyd,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .ge. ngxd+1 .and. ix .le. ngxd+ngx) then
c
              wvnum = wdata(ix-ngxd,iy-ngyd,ngx,ngy,xl,yl,
     &                      hx,hy,wreff,icase)
              alf   = fac    
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .eq. ngxd+ngx+1) then
c
              wvnum = wdata(ngx+1,iy-ngyd,ngx,ngy,xl,yl,
     &                      hx,hy,wreff,icase)
              alf   = fac 
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .ge. ngxd+ngx+2 .and. ix .le. ngxt) then
c
              wvnum = wdata(ngx+1,iy-ngyd,ngx,ngy,xl,yl,
     &                      hx,hy,wreff,icase)
              alf   = fac + const*((ix-(ngxd+ngx+1))*hx/xd)**2    
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            endif
          enddo
        elseif (iy .eq. ngyd+ngy+1) then
          do ix = 1,ngxt
            if (ix .ge. 1 .and. ix .le. ngxd-1) then
c
              wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngxd-ix)*hx/xd)**2   
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c 
            elseif (ix .eq. ngxd) then
c
              wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c
            elseif (ix .ge. ngxd+1 .and. ix .le. ngxd+ngx) then
c
              wvnum = wdata(ix-ngxd,ngy+1,ngx,ngy,xl,yl,
     &                      hx,hy,wreff,icase)
              alf   = fac             
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c    
            elseif (ix .eq. ngxd+ngx+1) then
c
              wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac   
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c     
            elseif (ix .ge. ngxd+ngx+2 .and. ix .le. ngxt) then
c
              wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ix-(ngxd+ngx+1))*hx/xd)**2    
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c      
            endif
          enddo
        elseif (iy .ge. ngyd+ngy+2 .and. iy .le. ngyt) then
          do ix = 1,ngxt
            if (ix .ge. 1 .and. ix .le. ngxd-1) then
              wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ngxd-ix)*hx/xd)**2
     &                    + const*((iy-(ngyd+ngy+1))*hy/yd)**2   
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c        
            elseif (ix .eq. ngxd) then
c            
              wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((iy-(ngyd+ngy+1))*hy/yd)**2    
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c          
            elseif (ix .ge. ngxd+1 .and. ix .le. ngxd+ngx) then
c
              wvnum = wdata(ix-ngxd,ngy+1,ngx,ngy,xl,yl,
     &                      hx,hy,wreff,icase)
              alf   = fac + const*((iy-(ngyd+ngy+1))*hy/yd)**2     
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c            
            elseif (ix .eq. ngxd+ngx+1) then
c
              wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((iy-(ngyd+ngy+1))*hy/yd)**2 
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c            
            elseif (ix .ge. ngxd+ngx+2 .and. ix .le. ngxt) then
c
              wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
              alf   = fac + const*((ix-(ngxd+ngx+1))*hx/xd)**2
     &                    + const*((iy-(ngyd+ngy+1))*hy/yd)**2 
              a(ix,iy,1) =  0.d0
              a(ix,iy,2) = -1.d0/hy2
              a(ix,iy,3) =  0.d0
              a(ix,iy,4) = -1.d0/hx2
              a(ix,iy,5) =  2.d0/hx2 + 2.d0/hy2 
     &                   - cmplx(c1,-(c2+alf))*wvnum**2  
              a(ix,iy,6) = -1.d0/hx2
              a(ix,iy,7) =  0.d0
              a(ix,iy,8) = -1.d0/hy2
              a(ix,iy,9) =  0.d0        
c           
            endif
          enddo
        endif
      enddo
c
c     Resolve the boundary conditions
c
c     West
      if (ibc(1,1) .eq. 0) then
         dxw = 0.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 1) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 0.d0
      elseif (ibc(1,1) .eq. 2) then
         dxw = 1.d0
         nxw = 1.d0
         sxw = 1.d0
      endif
c     East
      if (ibc(1,2) .eq. 0) then
         dxe = 0.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 1) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 0.d0
      elseif (ibc(1,2) .eq. 2) then
         dxe = 1.d0
         nxe = 1.d0
         sxe = 1.d0
      endif
c     South
      if (ibc(2,1) .eq. 0) then
         dys = 0.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 1) then
         dys = 1.d0
         nys = 1.d0
         sys = 0.d0
      elseif (ibc(2,1) .eq. 2) then
         dys = 1.d0
         nys = 1.d0
         sys = 1.d0
      endif
c     North
      if (ibc(2,2) .eq. 0) then
         dyn = 0.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 1) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 0.d0
      elseif (ibc(2,2) .eq. 2) then
         dyn = 1.d0
         nyn = 1.d0
         syn = 1.d0
      endif   
c
c
c     Include the boundary conditions
c
c     N:North    
      do ix = 2,ngxt-1
         if (ix .ge. 2 .and. ix .le. ngxd) then
            wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)    
	          omg   = cmplx(1.d0,-fac)*wvnum   
	          a(ix,ngyt,5) = a(ix,ngyt,5) 
     &                   + a(ix,ngyt,8)*dyn/(nyn + syn*cidd*omg*hy)
	          a(ix,ngyt,8) = 0.d0
	          wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
         elseif (ix .ge. ngxd+ngx+1 .and. ix .le. ngxt-1) then
            wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)    
	          omg   = cmplx(1.d0,-fac)*wvnum   
	          a(ix,ngyt,5) = a(ix,ngyt,5) 
     &                   + a(ix,ngyt,8)*dyn/(nyn + syn*cidd*omg*hy)
	          a(ix,ngyt,8) = 0.d0
         else
            wvnum = wdata(ix-ngxd,ngy+1,ngx,ngy,xl,yl,
     &                    hx,hy,wreff,icase)    
	          omg   = cmplx(1.d0,-fac)*wvnum   
	          a(ix,ngyt,5) = a(ix,ngyt,5) 
     &                   + a(ix,ngyt,8)*dyn/(nyn + syn*cidd*omg*hy)
	          a(ix,ngyt,8) = 0.d0
         endif
      enddo
c
c     S: South
      do ix = 2,ngxt-1
         if (ix .ge. 2 .and. ix .le. ngxd) then 
	          wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ix,1,5) = a(ix,1,5)
     &                + a(ix,1,2)*dys/(nys+ sys*cidd*omg*hy)
            a(ix,1,2) = 0.d0
         elseif (ix .ge. ngxd+ngx+1 .and. ix .le. ngxt-1) then
	          wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ix,1,5) = a(ix,1,5)
     &                + a(ix,1,2)*dys/(nys+ sys*cidd*omg*hy)
            a(ix,1,2) = 0.d0
         else
	          wvnum = wdata(ix-ngxd,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ix,1,5) = a(ix,1,5)
     &                + a(ix,1,2)*dys/(nys+ sys*cidd*omg*hy)
            a(ix,1,2) = 0.d0
         endif         
      enddo
c
c     W: West
      do iy = 2,ngyt-1
         if (iy .ge. 2 .and. iy .le. ngyd) then
	          wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(1,iy,5) = a(1,iy,5)
     &                + a(1,iy,4)*dxw/(nxw+ sxw*cidd*omg*hx)
            a(1,iy,4) = 0.d0
         elseif (iy .ge. ngyd+ngy+1 .and. iy .le. ngyt-1) then
	          wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(1,iy,5) = a(1,iy,5)
     &                + a(1,iy,4)*dxw/(nxw+ sxw*cidd*omg*hx)
            a(1,iy,4) = 0.d0
         else
	          wvnum = wdata(0,iy-ngyd,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(1,iy,5) = a(1,iy,5)
     &                + a(1,iy,4)*dxw/(nxw+ sxw*cidd*omg*hx)
            a(1,iy,4) = 0.d0
         endif
      enddo
c
c     E: East
      do iy = 2,ngyt-1
         if (iy .ge. 2 .and. iy .le. ngyd) then
	          wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ngxt,iy,5) = a(ngxt,iy,5)
     &                   + a(ngxt,iy,6)*dxe/(nxe+ sxe*cidd*omg*hx)
            a(ngxt,iy,6) = 0.d0
         elseif (iy .ge. ngyd+ngy+1 .and. iy .le. ngyt-1) then
	          wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,
     &                          icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ngxt,iy,5) = a(ngxt,iy,5)
     &                   + a(ngxt,iy,6)*dxe/(nxe+ sxe*cidd*omg*hx)
            a(ngxt,iy,6) = 0.d0
         else
	          wvnum = wdata(ngx+1,iy-ngyd,ngx,ngy,xl,yl,
     &                    hx,hy,wreff,icase)
	          omg   = cmplx(1.d0,-fac)*wvnum
	          a(ngxt,iy,5) = a(ngxt,iy,5)
     &                   + a(ngxt,iy,6)*dxe/(nxe+ sxe*cidd*omg*hx)
            a(ngxt,iy,6) = 0.d0
         endif
      enddo
c
c     Corner points
c
c     SW : Point(1,1)
      wvnum = wdata(0,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
      omg   = cmplx(1.d0,-fac)*wvnum
      a(1,1,5) = a(1,1,5) 
     &         + a(1,1,2)*dys/(nys + sys*cidd*omg*hy)
     &         + a(1,1,4)*dxw/(nxw + sxw*cidd*omg*hx) 
      a(1,1,2) = 0.d0
      a(1,1,4) = 0.d0      
c     SE : Point(ngxt,1)
      wvnum = wdata(ngx+1,0,ngx,ngy,xl,yl,hx,hy,wreff,icase)
      omg   = cmplx(1.d0,-fac)*wvnum
      a(ngxt,1,5) = a(ngxt,1,5) 
     &            + a(ngxt,1,2)*dys/(nys + sys*cidd*omg*hy)
     &            + a(ngxt,1,6)*dxe/(nxe + sxe*cidd*omg*hx)
      a(ngxt,1,2) = 0.d0
      a(ngxt,1,6) = 0.d0
c     NW : Point(1,ngyt)
      wvnum = wdata(0,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
      omg   = cmplx(1.d0,-fac)*wvnum
      a(1,ngyt,5) = a(1,ngyt,5) 
     &            + a(1,ngyt,8)*dyn/(nyn + syn*cidd*omg*hy)
     &            + a(1,ngyt,4)*dxw/(nxw + sxw*cidd*omg*hx) 
      a(1,ngyt,4) = 0.d0
      a(1,ngyt,8) = 0.d0
c     NE : Point(ngx,ngy)
      wvnum = wdata(ngx+1,ngy+1,ngx,ngy,xl,yl,hx,hy,wreff,icase)
      omg   = cmplx(1.d0,-fac)*wvnum
      a(ngxt,ngyt,5) = a(ngxt,ngyt,5) 
     &               + a(ngxt,ngyt,8)*dyn/(nyn + syn*cidd*omg*hy)
     &               + a(ngxt,ngyt,6)*dxe/(nxe + sxe*cidd*omg*hx)
      a(ngxt,ngyt,6) = 0.d0
      a(ngxt,ngyt,8) = 0.d0
c
      return
      end

      
                                                                      
                    
