      subroutine mkwitp2d(witp,a,
     &                    nx1,ny1,nx0,ny0,
     &                    ixs0,iys0,
     &                    ids,
     &                    neigx,neigy,
     &                    work0,work1,work2) 
c
c     This subroutine makes the weights of the prolongation mapping
c     from G(nx0,ny0) to G(nx1,ny1). 
c     wobst(nx1,ny1) indicates where is obstacles. 
c     If wobst(i,j) > 0.5, then (i,j) is included in a obstacle.
c
      implicit       none
      
      integer        ids,iw,k,l   
      integer        i,ix0,iy0,ix1,iy1 
      integer        ixs0,iys0,ixs1,iys1,ixs2,iys2,ixs3,iys3 
      integer        nx0,ny0,nx1,ny1 
      integer        neigx(2),neigy(2) 
      complex*16     witp(nx1,ny1,4) 
      complex*16     a(nx1,ny1,-1:1,-1:1) 
      complex*16     work0(0:nx1+1,0:ny1+1) 
      complex*16     work1(0:nx1+1,0:ny1+1) 
      complex*16     work2(0:nx1+1,0:ny1+1) 
      complex*16     czero,cone,chalf,cidd 
      complex*16     d,d1,d2,d3,d4,w1,w2,w3,w4  
      real*8         epsvs,dhalf,wtol,da,db,dc,dw,de,dn,ds,sigm,
     &               ww,we,wn,ws, sum,c1,c2,w1w,w1e,w1s,w1n,w2s   
      real*8         dae,daw,das,dan 
      
      parameter      (czero = (0.d0,0.d0),
     &                cone  = (1.d0,0.d0),
     &                chalf = (0.5d0,0.d0),
     &                cidd  = (0.d0,1.d0),
     &                dhalf = 0.5d0,
     &                epsvs = 1.d-32,
     &                wtol  = 2.d0)
c
      if ((nx1 .lt. 1) .or. (ny1 .lt. 1)) return
c
c     initialization
c
      do iy1 = 1,ny1
         do ix1 = 1,nx1
            witp(ix1,iy1,1) = czero
            witp(ix1,iy1,2) = czero
            witp(ix1,iy1,3) = czero
            witp(ix1,iy1,4) = czero
         enddo
      enddo
C
      ixs1 = mod(ixs0,2)+1
      iys1 = iys0
      ixs2 = ixs0
      iys2 = mod(iys0,2)+1
      ixs3 = mod(ixs0,2)+1
      iys3 = mod(iys0,2)+1 
c
c     mark on the coarser grid points except for the obstraction
c
      do iy1 = 0, ny1+1
         do ix1 = 0, nx1+1
            work0(ix1,iy1) = czero
         enddo
      enddo
      
      do iy0 = 1, ny0
         iy1 = iys0 + iy0*2 -2
         do ix0 = 1, nx0
            ix1 = ixs0 + ix0*2 -2
c            if (wobst(ix1,iy1) .lt. dhalf) work0(ix1, iy1) = cone
            work0(ix1, iy1) = cone
         enddo
      enddo
      
      if (ids .eq. 0) then
c
c     on grid points 1
c
        do iy1 = iys1, ny1, 2
          do ix1 = ixs1, nx1, 2
            d1 = czero
            d2 = czero
            if (abs(work0(ix1-1,iy1)) .gt. dhalf) d1 = cone
            if (abs(work0(ix1+1,iy1)) .gt. dhalf) d2 = cone
            d = d1 + d2
            if (abs(d) .gt. dhalf) then
               witp(ix1,iy1,1) = d1/d 
               witp(ix1,iy1,2) = d2/d 
            endif
          enddo
        enddo
c
c     on grid points 2
c
        do iy1 = iys2, ny1, 2
          do ix1 = ixs2, nx1, 2
            d1 = czero
            d2 = czero
            if (abs(work0(ix1,iy1-1)).gt.dhalf) d1 = cone
            if (abs(work0(ix1,iy1+1)).gt.dhalf) d2 = cone
            d = d1 + d2
            if (abs(d) .gt. dhalf) then
               witp(ix1,iy1,1) = d1/d 
               witp(ix1,iy1,2) = d2/d 
            endif
         enddo
       enddo
c
c     on grid points 3
c
       do iy1 = iys3, ny1, 2
         do ix1 = ixs3, nx1, 2
            d1 = czero
            d2 = czero
            d3 = czero
            d4 = czero
            if (abs(work0(ix1-1,iy1-1)) .gt. dhalf) d1 = cone
            if (abs(work0(ix1+1,iy1-1)) .gt. dhalf) d2 = cone
            if (abs(work0(ix1-1,iy1+1)) .gt. dhalf) d3 = cone
            if (abs(work0(ix1+1,iy1+1)) .gt. dhalf) d4 = cone
            d = d1 + d2 + d3 + d4 
            if (abs(d) .gt. dhalf) then
               witp(ix1,iy1,1) = d1/d 
               witp(ix1,iy1,2) = d2/d 
               witp(ix1,iy1,3) = d3/d 
               witp(ix1,iy1,4) = d4/d 
            endif
          enddo
        enddo

      elseif (ids .eq. 1) then
c
c     Matrix dependent
c
c        on grid points 1
c
         do iy1 = iys1, ny1, 2
            do ix1 = ixs1, nx1, 2
               d1 = a(ix1,iy1,-1,-1)+a(ix1,iy1,-1,0)+a(ix1,iy1,-1,1)
               d2 = a(ix1,iy1,1,-1) +a(ix1,iy1,1,0) +a(ix1,iy1,1,1)
c
               d1 = work0(ix1-1,iy1)*d1
               d2 = work0(ix1+1,iy1)*d2
c
               d = -(a(ix1,iy1,0,0)+a(ix1,iy1, 0,-1)+a(ix1,iy1, 0, 1))
c            
               iw = 0
               if (abs(d) .gt. epsvs) then
                  w1 = d1/d
                  w2 = d2/d
                  if (abs(w1)+abs(w2) .lt. wtol) then
                     witp(ix1,iy1,1) = w1
                     witp(ix1,iy1,2) = w2
                     iw = 1
                  endif
               endif
               if (iw.eq.0) then
                  d1 = czero
                  d2 = czero
                  if (abs(work0(ix1-1,iy1)).gt.dhalf) d1 = cone
                  if (abs(work0(ix1+1,iy1)).gt.dhalf) d2 = cone
                  d = d1 + d2
                  if (abs(d) .gt. dhalf) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                  endif
               endif
            enddo
         enddo
c
c        on grid points 2
c
         do iy1 = iys2, ny1, 2
            do ix1 = ixs2, nx1, 2
               d1 = a(ix1,iy1,-1,-1)+ a(ix1,iy1,0,-1) + a(ix1,iy1,1,-1)
               d2 = a(ix1,iy1,-1, 1)+ a(ix1,iy1,0, 1) + a(ix1,iy1,1, 1)
c
               d1 = work0(ix1,iy1-1)*d1
               d2 = work0(ix1,iy1+1)*d2
c
c              d =  d1 + d2
               d = -(a(ix1,iy1,0,0)+a(ix1,iy1,-1, 0)+a(ix1,iy1, 1, 0))
               iw = 0
               if (abs(d).gt.epsvs) then
                  w1 = d1/d
                  w2 = d2/d
                  if (abs(w1)+abs(w2).lt.wtol) then
                     witp(ix1,iy1,1) = w1
                     witp(ix1,iy1,2) = w2
                     iw = 1
                  endif
               endif
               if (iw.eq.0) then
                  d1 = czero
                  d2 = czero
                  if (abs(work0(ix1,iy1-1)).gt.dhalf) d1 = cone
                  if (abs(work0(ix1,iy1+1)).gt.dhalf) d2 = cone
                  d = d1 + d2
                  if (abs(d) .gt. dhalf) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                  endif
               endif
            enddo
         enddo
c
         do iy1 = 0, ny1+1
            do ix1 = 0, nx1+1
               work1(ix1,iy1) = czero
               work2(ix1,iy1) = czero
            enddo
         enddo
         do iy1 = 1, ny1
            do ix1 = 1, nx1
               work1(ix1,iy1) = witp(ix1,iy1,1)
               work2(ix1,iy1) = witp(ix1,iy1,2)
            enddo
         enddo
c
c     on grid points 3
c
         do iy1 = iys3, ny1, 2
            do ix1 = ixs3, nx1, 2
               d1 = a(ix1,iy1,-1,-1)
     &            + a(ix1,iy1,0,-1)*work1(ix1,iy1-1)
     &            + a(ix1,iy1,-1,0)*work1(ix1-1,iy1)
               d2 = a(ix1,iy1,1,-1)
     &            + a(ix1,iy1,0,-1)*work2(ix1,iy1-1)
     &            + a(ix1,iy1,1,0)*work1(ix1+1,iy1)
               d3 = a(ix1,iy1,-1,1)
     &            + a(ix1,iy1,0,1)*work1(ix1,iy1+1)
     &            + a(ix1,iy1,-1,0)*work2(ix1-1,iy1)
               d4 = a(ix1,iy1,1,1)
     &            + a(ix1,iy1,0,1)*work2(ix1,iy1+1)
     &            + a(ix1,iy1,1,0)*work2(ix1+1,iy1)
c
               d1 = work0(ix1-1,iy1-1)*d1
               d2 = work0(ix1+1,iy1-1)*d2
               d3 = work0(ix1-1,iy1+1)*d3
               d4 = work0(ix1+1,iy1+1)*d4
c
c            d = d1 + d2 + d3 + d4
               d = -a(ix1,iy1,0,0)
               iw = 0
               if (abs(d) .gt. epsvs) then
                  w1 = d1/d
                  w2 = d2/d
                  w3 = d3/d
                  w4 = d4/d
                  if (abs(w1)+abs(w2)+abs(w3)+abs(w4)
     &               .lt.wtol) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                     witp(ix1,iy1,3) = d3/d
                     witp(ix1,iy1,4) = d4/d
                     iw = 1
                  endif
               endif
               if (iw.eq.0) then
                  d1 = czero
                  d2 = czero
                  d3 = czero
                  d4 = czero
                  if (abs(work0(ix1-1,iy1-1)).gt.dhalf) d1 = cone
                  if (abs(work0(ix1+1,iy1-1)).gt.dhalf) d2 = cone
                  if (abs(work0(ix1-1,iy1+1)).gt.dhalf) d3 = cone
                  if (abs(work0(ix1+1,iy1+1)).gt.dhalf) d4 = cone
                  d = d1 + d2 + d3 + d4
                  if (abs(d) .gt. dhalf) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                     witp(ix1,iy1,3) = d3/d
                     witp(ix1,iy1,4) = d4/d
                  endif
               endif
            enddo
         enddo
c
      elseif (ids .eq. 2) then
c
c     on grid points 1
c
         do iy1 = iys1,ny1,2
            do ix1 = ixs1,nx1,2
c
               da = cdabs(a(ix1,iy1,-1,-1) + a(ix1,iy1,-1, 0) +
     &                    a(ix1,iy1,-1, 1))
               db = cdabs(a(ix1,iy1,-1,-1))
               dc = cdabs(a(ix1,iy1,-1, 1))
               dw = max(da,db,dc)
               daw = da;
c
               da = cdabs(a(ix1,iy1, 1,-1) + a(ix1,iy1, 1, 0) +
     &                    a(ix1,iy1, 1, 1))
               db = cdabs(a(ix1,iy1, 1,-1))
               dc = cdabs(a(ix1,iy1, 1, 1))
               de = max(da,db,dc)
               dae = da;
c               if (dae-daw .ne. 0.d0) write(*,*) dae-daw
c
               da = cdabs(a(ix1,iy1,-1, 1) + a(ix1,iy1, 0, 1) +
     &                    a(ix1,iy1, 1, 1))
               db = cdabs(a(ix1,iy1,-1, 1))
               dc = cdabs(a(ix1,iy1, 1, 1))
               dn = max(da,db,dc)
               dan = da;
c
               da = cdabs(a(ix1,iy1,-1,-1) + a(ix1,iy1, 0,-1) +
     &                    a(ix1,iy1, 1,-1))
               db = cdabs(a(ix1,iy1,-1,-1))
               dc = cdabs(a(ix1,iy1, 1,-1))
               ds = max(da,db,dc)
               das = da;
c               if (das-dan .ne. 0.d0) write(*,*) das-dan
               
c
               sum = 0.d0
               do l = -1, 1
                  do k = -1, 1
                    sum = sum + cdabs(a(ix1,iy1,k,l))
                  enddo
               enddo
               sigm = 0.5d0*dmin1(1.d0,cdabs(1.d0-sum/a(ix1,iy1, 0, 0)))
               c1 = cdabs(a(ix1,iy1, 1,-1)) + cdabs(a(ix1,iy1, 1, 0)) + 
     &              cdabs(a(ix1,iy1, 1, 1)) - cdabs(a(ix1,iy1, 1,-1)) -
     &              cdabs(a(ix1,iy1, 1, 0)) - cdabs(a(ix1,iy1, 1, 1)) -
     &              cdabs(a(ix1,iy1,-1,-1)) - cdabs(a(ix1,iy1,-1, 0)) -
     &              cdabs(a(ix1,iy1,-1, 1)) + cdabs(a(ix1,iy1,-1,-1)) +
     &              cdabs(a(ix1,iy1,-1, 0)) + cdabs(a(ix1,iy1,-1, 1))
c               c1 = 0.d0
c               if (c1 .ge. 1e-12) write(*,*) c1
c
               if (dw+de .gt. epsvs) then
                  w1w = sigm*(1.d0 + (dw-de)/(dw+de) + c1/(dw+de+ds+dn))
                  w1e = 2.d0*sigm - w1w
c
                  w1w = max(0.d0,w1w)
                  w1e = max(0.d0,w1e)
                  w1 = min(2.d0*sigm,w1w)
                  w2 = min(2.d0*sigm,w1e)
c
c    Obstacle
c
                  w1 = work0(ix1-1,iy1)*w1
                  w2 = work0(ix1+1,iy1)*w2
c
                  witp(ix1,iy1,1) = w1
                  witp(ix1,iy1,2) = w2
               else
                  witp(ix1,iy1,1) = 0.d0
                  witp(ix1,iy1,2) = 0.d0
               endif
            enddo
         enddo
c
c     on grid points 2
c
         do iy1 = iys2, ny1, 2
            do ix1 = ixs2, nx1, 2
c
               da = cdabs(a(ix1,iy1,-1,-1) + a(ix1,iy1,-1, 0) +
     &                    a(ix1,iy1,-1, 1))
               db = cdabs(a(ix1,iy1,-1,-1))
               dc = cdabs(a(ix1,iy1,-1, 1))
               dw = max(da,db,dc)
c
               da = cdabs(a(ix1,iy1, 1,-1) + a(ix1,iy1, 1, 0) +
     &                    a(ix1,iy1, 1, 1))
               db = cdabs(a(ix1,iy1, 1,-1))
               dc = cdabs(a(ix1,iy1, 1, 1))
               de = max(da,db,dc)
c
               da = cdabs(a(ix1,iy1,-1, 1) + a(ix1,iy1, 0, 1) +
     &                    a(ix1,iy1, 1, 1))
               db = cdabs(a(ix1,iy1,-1, 1))
               dc = cdabs(a(ix1,iy1, 1, 1))
               dn = max(da,db,dc)
c
               da = cdabs(a(ix1,iy1,-1,-1) + a(ix1,iy1, 0,-1) +
     &                  a(ix1,iy1, 1,-1))
               db = cdabs(a(ix1,iy1,-1,-1))
               dc = cdabs(a(ix1,iy1, 1,-1))
               ds = max(da,db,dc)
c
               sum = 0.d0
               do l = -1, 1
                  do k = -1, 1
                    sum = sum + cdabs(a(ix1,iy1,k,l))
                  enddo
               enddo
               sigm = 0.5d0*dmin1(1.d0,cdabs(1.d0-sum/a(ix1,iy1, 0, 0)))
               c2 = cdabs(a(ix1,iy1,-1, 1)) + cdabs(a(ix1,iy1, 0, 1)) +
     &              cdabs(a(ix1,iy1, 1, 1)) - cdabs(a(ix1,iy1,-1, 1)) -
     &              cdabs(a(ix1,iy1, 0, 1)) - cdabs(a(ix1,iy1, 1, 1)) -
     &              cdabs(a(ix1,iy1,-1,-1)) - cdabs(a(ix1,iy1, 0,-1)) -
     &              cdabs(a(ix1,iy1, 1,-1)) + cdabs(a(ix1,iy1,-1,-1)) +
     &              cdabs(a(ix1,iy1, 0,-1)) + cdabs(a(ix1,iy1, 1,-1))
c               c2 = 0.d0
c               if (c2 .ge. 1.e-12) write(*,*) c2
c
               if (ds+dn.gt.epsvs) then
                  w1s = sigm*(1.d0 + (ds-dn)/(ds+dn) + c2/(dw+de+ds+dn))
                  w1n = 2.d0*sigm - w1s
c
                  w1s = max(0.d0,w1s)
                  w2s = max(0.d0,w1n) 
                  w1 = min(2.d0*sigm,w1s)
                  w2 = min(2.d0*sigm,w1n)
c     
c    Obstacle
c
                  w1 = work0(ix1,iy1-1)*w1
                  w2 = work0(ix1,iy1+1)*w2
c     
                  witp(ix1,iy1,1) = w1
                  witp(ix1,iy1,2) = w2
               else
                  witp(ix1,iy1,1) = 0.d0
                  witp(ix1,iy1,2) = 0.d0
               endif
            enddo
         enddo
c
         do iy1 = 0, ny1+1
            do ix1 = 0, nx1+1
               work1(ix1,iy1) = 0.d0
               work2(ix1,iy1) = 0.d0
            enddo
         enddo
         do iy1 = 1, ny1
            do ix1 = 1, nx1
               work1(ix1,iy1) = witp(ix1,iy1,1)
               work2(ix1,iy1) = witp(ix1,iy1,2)
            enddo
         enddo
c
c     on grid points 3
c
         do iy1 = iys3, ny1, 2
            do ix1 = ixs3, nx1, 2
               d1 = a(ix1,iy1,-1,-1)
     &            + a(ix1,iy1,0,-1)*work1(ix1,iy1-1)
     &            + a(ix1,iy1,-1,0)*work1(ix1-1,iy1)
               d2 = a(ix1,iy1,1,-1)
     &            + a(ix1,iy1,0,-1)*work2(ix1,iy1-1)
     &            + a(ix1,iy1,1,0)*work1(ix1+1,iy1)
               d3 = a(ix1,iy1,-1,1)
     &            + a(ix1,iy1,0,1)*work1(ix1,iy1+1)
     &            + a(ix1,iy1,-1,0)*work2(ix1-1,iy1)
               d4 = a(ix1,iy1,1,1)
     &            + a(ix1,iy1,0,1)*work2(ix1,iy1+1)
     &            + a(ix1,iy1,1,0)*work2(ix1+1,iy1)
c
               d1 = work0(ix1-1,iy1-1)*d1
               d2 = work0(ix1+1,iy1-1)*d2
               d3 = work0(ix1-1,iy1+1)*d3
               d4 = work0(ix1+1,iy1+1)*d4
c
               d = -a(ix1,iy1,0,0)
               iw = 0
               if (abs(d) .gt. epsvs) then
                  w1 = d1/d
                  w2 = d2/d
                  w3 = d3/d
                  w4 = d4/d
                  if (abs(w1)+abs(w2)+abs(w3)+abs(w4) .lt. wtol) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                     witp(ix1,iy1,3) = d3/d
                     witp(ix1,iy1,4) = d4/d
                     iw = 1
                  endif
               endif
               if (iw.eq.0) then
                  d1 = 0.d0
                  d2 = 0.d0
                  d3 = 0.d0
                  d4 = 0.d0
                  if (abs(work0(ix1-1,iy1-1)).gt.dhalf) d1 = cone
                  if (abs(work0(ix1+1,iy1-1)).gt.dhalf) d2 = cone 
                  if (abs(work0(ix1-1,iy1+1)).gt.dhalf) d3 = cone 
                  if (abs(work0(ix1+1,iy1+1)).gt.dhalf) d4 = cone 
                  d = d1 + d2 + d3 + d4
                  if (abs(d) .gt. dhalf) then
                     witp(ix1,iy1,1) = d1/d
                     witp(ix1,iy1,2) = d2/d
                     witp(ix1,iy1,3) = d3/d
                     witp(ix1,iy1,4) = d4/d
                  endif
               endif
            enddo
         enddo
c        
      endif
c
      return
      end
