      subroutine pjac2f(a,dinv,w,b,
     &                  nx,ny,xl,yl,
     &                  nr,omega, 
     &                  ix0,iy0,work)
c
c     4-color SOR iteration for 2-dim 5-point finite difference used  in
c     the finest grid.
c
c     Variable definition:
c        NGX,NGY : grid size
c        w       : initial value and solution
c        B       : righthand side
c        OMEGA   : acceleration parameter (for Gauss-Seidel, omega=1)
c        nr      : number of iterations 
c
      implicit       none
      
      integer        nx,ny,nr,ix0,iy0 
      integer        ix,iy,ixc,iyc,ixst,iyst 
      integer        i,j,it   
      complex*16     a(nx,ny,-1:1,-1:1),dinv(nx,ny) 
      complex*16     b(nx,ny),w(nx,ny) 
      complex*16     work(0:nx+1,0:ny+1) 
      complex*16     d 
      real*8         xl,yl  
      complex*16     omega 

c
c     Prepare the temporary variable used during iteration. The varia-
c     ble is work.

      do iy = 0,ny+1
         if (iy .eq. 0 .or. iy .eq. ny+1) then
            do ix = 0,nx+1
               work(ix,iy) = (0.d0,0.d0)
            enddo
         else
            do ix = 0,nx+1
               if (ix .eq. 0 .or. ix .eq. nx+1) then
                  work(ix,iy) = (0.d0,0.d0)
               else
                  work(ix,iy) = w(ix,iy)
               endif
            enddo
         endif
      enddo
c
c     compute the color number of the bottom line : ic0
c
      do it = 1,nr
c              Start j sweep
               do j = 1,ny
c                 Start i sweep
                  do i = 1,nx
                     d = omega*dinv(i,j)*(b(i,j) 
     &                 - a(i,j, 0,-1)*work(i,j-1)
     &                 - a(i,j, 0, 1)*work(i,j+1)
     &                 - a(i,j,-1, 0)*work(i-1,j)
     &                 - a(i,j, 1, 0)*work(i+1,j)
     &                 - a(i,j, 0, 0)*work(i,j))
                     w(i,j) = w(i,j) +d
c                    End i sweep
                  enddo
c              End j sweep
               enddo
      enddo
      
      return
      end

c***********************************************************************

      subroutine pjac2c(a,dinv,w,b,
     &                  nx,ny,xl,yl,
     &                  nr,omega, 
     &                  ix0,iy0,work)
c
c     4-color SOR iteration for 2-dim 9 point finite difference used  on
c     coarser grid level.
c
c     Variable definition:
c        NGX,NGY : grid size
c        w       : initial value and solution
c        B       : righthand side
c        OMEGA   : acceleration parameter (for Gauss-Seidel, omega=1)
c        nr      : number of iterations 
c
      implicit       none
      
      integer        nx,ny,nr,ix0,iy0 
      integer        ix,iy,ixc,iyc,ixst,iyst 
      integer        i,j,it  
      complex*16     a(nx,ny,-1:1,-1:1),dinv(nx,ny) 
      complex*16     b(nx,ny),w(nx,ny) 
      complex*16     work(0:nx+1,0:ny+1) 
      complex*16     d 
      real*8         xl,yl 
      complex*16     omega 

c
c     Prepare the temporary variable used during iteration. The varia-
c     ble is work.

      do iy = 0,ny+1
         if (iy .eq. 0 .or. iy .eq. ny+1) then
            do ix = 0,nx+1
               work(ix,iy) = (0.d0,0.d0)
            enddo
         else
            do ix = 0,nx+1
               if (ix .eq. 0 .or. ix .eq. nx+1) then
                  work(ix,iy) = (0.d0,0.d0)
               else
                  work(ix,iy) = w(ix,iy)
               endif
            enddo
         endif
      enddo
c
c     compute the color number of the bottom line : ic0
c
      do it = 1,nr
c              Start j sweep
               do j = 1,ny
c                 Start i sweep
                  do i = 1,nx    
                     d = omega*dinv(i,j)*(b(i,j) 
     &                 - a(i,j, 0,-1)*work(i,j-1)
     &                 - a(i,j,-1, 0)*work(i-1,j)
     &                 - a(i,j, 0, 0)*work(i,j)
     &                 - a(i,j, 1, 0)*work(i+1,j)
     &                 - a(i,j, 0, 1)*work(i,j+1))
                     w(i,j) = w(i,j) + d
c                    End i sweep
                  enddo
c              End j sweep
               enddo
      enddo
      
      return
      end
