      subroutine prntdata(ia,n,nxpml,nypml,xl,yl,inump,itdata,time,x,
     &                    res)
      
      
C     This subroutine serves as the driver to print convergence data in-
C     to files for post-processing. Two files are generated: the numeri-
C     cal performance data and the solution. The  numerical  performance
C     data is printed in NUMDATA.OUT, while the solution  is  saved in
C     SOLUTION.OUT.

      IMPLICIT NONE
      
      integer        ia(2),inump(7),niter,iprnt(1),igrid 
      integer        itdata(5) 
      real*8         res(2)   
      integer        i,ix,iy,ix0,iy0,ix1,iy1,n,ngx,ngy,ngxp,ngyp 
      integer        nxpml,nypml  
      complex*16     x(n) 
      real*8         xd,yd,xl,yl,dx,dy  
      real*8         time(5) 
      character*40   a1,a2,a3 

      NGX = IA(1)
      NGY = IA(2)
      
      if (INUMP(1) .EQ. 1) then
         a1 = 'BICGSTAB'
      elseif (INUMP(1) .EQ. 2) then
         a1 = 'COCG'
      elseif (INUMP(1) .EQ. 3) then
         a1 = 'GCR'
      elseif (inump(1) .eq. 4) then
         a1 = 'GMRES'
      elseif (inump(1) .eq. 5) then
         a1 = 'GMRESR' 
      elseif (inump(1) .eq. 6) then
         a1 = 'Multigrid'
      elseif (inump(1) .eq. 7) then
         a1 = 'BiCGSTAB(2)'
      endif
      
      IF (INUMP(4) .EQ. 0) THEN
         A2 = 'No preconditioner is used'
      ELSEIF (INUMP(4) .EQ. 1) THEN
         A2 = 'Original Helmholtz'
      ELSEIF (INUMP(4) .EQ. 2) THEN
         A2 = 'Shifted-Laplace'
      elseif (inump(4) .eq. 3) then
         a2 = 'Monga-Made preconditioner'
      ENDIF
      
      IF (INUMP(5) .EQ. 1) THEN
         A3 = 'ILU(0)'
      ELSEIF (INUMP(5) .EQ. 2) THEN
         A3 = 'ILU(1)'
      ELSEIF (INUMP(5) .EQ. 3) THEN
         A3 = 'Inner iteration'
      ELSEIF (INUMP(5) .EQ. 4) THEN
         A3 = 'Multigrid'
      ENDIF              

C     ******************************************************************
C     Print numerical performance
    
      WRITE(*,*)  '2D Helmholtz problem'
      WRITE(*,*)
      WRITE(*,*)  '---------------------------------------------'
      WRITE(*,*)
      WRITE(*,*)  'Grid Data: '
      WRITE(*,100)'... No. of gridpoints in x-direction   : ',NGX
      WRITE(*,100)'... No. of gridpoints in y-direction   : ',NGY
        
      WRITE(*,*)  '---------------------------------------------'
      WRITE(*,*)
      WRITE(*,*)  'Solution method: '
      WRITE(*,110)'... Type of solver                     : ',a1
      WRITE(*,110)'... Preconditioner                     : ',a2
      WRITE(*,110)'... Method to solve the preconditioner : ',a3
      WRITE(*,100)'... Number of iteration                : ',itdata(1)
      write(*,100)'... Number of mat-vec multiplications  : ',itdata(2)
      write(*,120)'... || res || at convergence           : ',res(1)
      write(*,120)'... || res || relative at convergence  : ',res(2)
      WRITE(*,*)
      WRITE(*,*)  'Time data (in seconds)'
      WRITE(*,120)'... Time spent for initialization      : ',TIME(1)
      WRITE(*,120)'... Time for the main iteration        : ',TIME(4)
      WRITE(*,120)'... Time for solve A*x                 : ',TIME(2)
      WRITE(*,120)'... Time for preconditioner solve      : ',TIME(3)
      WRITE(*,120)'... Total time to convergence          : ',TIME(5)
      WRITE(*,*)
      WRITE(*,*) '---------------------------------------------'
      
      OPEN(UNIT=30,FILE='numdata.out')
      
      WRITE(30,*)  '2D Helmholtz problem'

      WRITE(30,*)
      WRITE(30,*)  '---------------------------------------------'
      WRITE(30,*)
      WRITE(30,*)  'Grid Data: '
      WRITE(30,100)'... No. of gridpoints in x-direction   : ',NGX
      WRITE(30,100)'... No. of gridpoints in y-direction   : ',NGY
          
      WRITE(30,*)  '---------------------------------------------'
      WRITE(30,*)
      WRITE(30,*)  'Solution method: '
      WRITE(30,110)'... Type of solver                     : ',a1
      WRITE(30,110)'... Preconditioner                     : ',a2
      WRITE(30,110)'... Method to solve the preconditioner : ',a3
      WRITE(30,100)'... Number of iteration                : ',itdata(1)
      write(30,100)'... Number of mat-vec multiplications  : ',itdata(2)
      write(30,120)'... || res || at convergence           : ',res(1)
      write(30,120)'... || res || relative at convergence  : ',res(2)
      WRITE(30,*)
      WRITE(30,*)  'Time data (in seconds)'
      WRITE(30,120)'... Time spent for initialization      : ',TIME(1)
      WRITE(30,120)'... Time for the main iteration        : ',TIME(4)
      WRITE(30,120)'... Time for solve A*x                 : ',TIME(2)
      WRITE(30,120)'... Time for preconditioner solve      : ',TIME(3)
      WRITE(30,120)'... Total time to convergence          : ',TIME(5)
      WRITE(30,*)
      WRITE(30,*) '---------------------------------------------'
      
      CLOSE(30)
      
C     ******************************************************************
C     Print the solution
      
      igrid = itdata(3)
      if (itdata(4) .ne. 0) then
        if (igrid .eq. 0) then
           dx = xl/float(ngx+1)
           dy = yl/float(ngy+1)
           ix0 = 1
           iy0 = 1
           ix1 = ngx
           iy1 = ngy
           open(unit=31,file='solution.out')
           i = 1
           do iy = iy0,iy1 
              yd = dy*float(iy)
              do ix = ix0,ix1 
                 xd = dx*float(ix)
                 write(31,200) xd,yd,realpart(x((iy-1)*ngx+ix)),
     &                         imagpart(x((iy-1)*ngx+ix))
              enddo
           enddo
           close(31)
        elseif (igrid .eq. 1) then
           dx = xl/float(ngx-1)
           dy = yl/float(ngy-1)
           ix0 = 1
           iy0 = 1
           ix1 = ngx-2
           iy1 = ngy-2
           write(*,*) ix0,ix1,iy0,iy1
           open(unit=31,file='solution.out')
           i = 1
           do iy = iy0,iy1 
              yd = dy*float(iy)
              do ix = ix0,ix1 
                 xd = dx*float(ix)
                 write(31,200) xd,yd,realpart(x(iy*ngx+ix+1)),
     &                         imagpart(x(iy*ngx+ix+1))
              enddo
           enddo
           close(31)
        elseif (igrid .eq. 2) then
           ngxp = ngx - 2*nxpml
           ngyp = ngy - 2*nypml
           
           dx = xl/float(ngxp+1)
           dy = yl/float(ngyp+1)
           write(*,*) ngxp,ngyp,dx,dy 
           ix0 =   nxpml+1
           iy0 =   nypml+1
           ix1 =   ngxp + nxpml
           iy1 =   ngyp + nypml
           open(unit=31,file='solution.out')
           i = 1
           do iy = iy0,iy1 
              yd = dy*float(iy-nypml)
              do ix = ix0,ix1 
                 xd = dx*float(ix-nxpml)
                 write(31,200) xd,yd,realpart(x((iy-1)*ngx+ix)),
     &                         imagpart(x((iy-1)*ngx+ix))
                 i = i + 1
              enddo
           enddo
           close(31)
        endif   
      endif
c      
  100 format(' ',A,I5)
  110 format(' ',A,A30)
  120 format(' ',A,F18.9)
  200 format('  ',D14.8,'  ',D14.8,'  ',D14.8,'  ',D14.8)
  300 format('  ',I8,' ',D14.8,'  ',D14.8,'  ',D14.8,'  ',D14.8)    
      return
      end
        
